#include "Downloader.h"


#ifdef _WIN32							// Windows

static void sleep(unsigned sec)
{
	Sleep(1000 * sec);
}

static int close(SOCKET sockfd)
{
	return closesocket(sockfd);
}

static Error initWinsockDll()
{
	WORD wVersionRequested;
	WSADATA wsaData;

	wVersionRequested = MAKEWORD(2, 2); 

	if (WSAStartup(wVersionRequested, &wsaData)){
		fprintf(stderr, "WSAStartup failed\n");
		return ERROR_INITWINSOCKDLL;
	}

	if (LOBYTE(wsaData.wVersion) != 2 ||
        HIBYTE(wsaData.wVersion) != 2 ){
		fprintf(stderr, "LOBYTE = %d\n", LOBYTE(wsaData.wVersion));
		fprintf(stderr, "HIBYTE = %d\n", HIBYTE(wsaData.wVersion));

		WSACleanup();
		return ERROR_INITWINSOCKDLL; 
	}
	
	return NOT_ERROR;
}

#endif


// Update a given URL string for downloading.
// Input:	a string that is in the exact form of BASE_URL.
// Return:	a pointer to the string (equals the given pointer).
// Note:	the given string is being changed.
static char* updateURL(char base_url[])
{
	char* ptr;
	const time_t now = time(NULL);
	struct tm* date = gmtime(&now);

	assert(date);

	date->tm_hour = ((int)(date->tm_hour / download_cycle)) *
					download_cycle;
	date->tm_min = 0;
	
	ptr = (char *)strstr(base_url, "0000");
	assert(ptr);

	strftime(ptr, strlen(ptr), "%Y.%m/bview.%Y%m%d.%H%M.gz", date);

	return base_url;
}


// Download a url of a host via a port, and write it to a file.
// Return: an Error index.
static Error download(FILE* downloaded, char* host,
					  unsigned short port, char* url)
{
	unsigned bytes, size_of_char = sizeof(char);
	char buffer[BUFFER_SIZE];
	SOCKET sockfd;
	struct sockaddr_in addr;
	struct hostent *h_ent;

	if ((h_ent = gethostbyname(host)) == NULL){
		fprintf(stderr, "\nERROR:\nUnknown host.\n");
		return ERROR_UNKNOWN_HOST;
	}

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	memcpy((char *)&(addr.sin_addr.s_addr),
			h_ent->h_addr_list[0], h_ent->h_length);

	if ((sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0){
		fprintf(stderr, "\nERROR:\nCannot open socket.\n");
		return ERROR_OPEN_SOCKET;
	}

	if (connect(sockfd, (struct sockaddr *)&addr, sizeof(addr)) < 0){
		fprintf(stderr, "\nERROR:\nConnection failed.\n");
		return ERROR_CONNECTION_FAILED;
	}

	sprintf(buffer, "GET %s HTTP/1.0\r\nHost: %s\r\n\r\n", url, host);

	if (send(sockfd, buffer, strlen(buffer), 0) <0 ){
		fprintf(stderr,
			"\nERROR:\nFailed sending request to host.\n");
		close(sockfd);
		return ERROR_SEND_REQUEST;
    }

	memset(buffer, 0, BUFFER_SIZE);

	// get response and ignore the HTTP header
	if ((bytes = recv(sockfd, buffer, BUFFER_SIZE, 0)) > 0){
		char *body = strrchr(buffer, '\n');
		unsigned len;

		assert(body);
		body[0] = '\0';	// terminate the HTTP header
		body++;
		len = bytes - strlen(buffer) - 1;

		if (fwrite(body, size_of_char, len, downloaded) != len){
			fprintf(stderr,
				"\nERROR:\nFailed writing to file \"%s\".\n", downloaded);
			close(sockfd);
			return ERROR_WRITE_FILE;
		}
	}

	// continue downloading the file
	while ((bytes = recv(sockfd, buffer, BUFFER_SIZE, 0)) > 0){
		if (fwrite(buffer, size_of_char, bytes, downloaded) != bytes){
			fprintf(stderr,
				"\nERROR:\nFailed writing to file \"%s\".\n", downloaded);
			close(sockfd);
			return ERROR_WRITE_FILE;
		}
	}

	if (close(sockfd) < 0){
		fprintf(stderr,
			"\nERROR:\nFailed in closing the connection.\n");
		return ERROR_CLOSE_SOCKET;
	}

	return NOT_ERROR;
}


// the global io file pointer of the downloader
static FILE *io_file = NULL;

// The downloader thread's handler for the SIGTERM signal.
static void downloaderExit(int error)
{
	if (io_file)
		fclose(io_file);

	pthread_mutex_unlock(&trie_locker);

	pthread_cond_broadcast(&startup_event);

#ifdef _WIN32
	raise(SIGUSR1);
#else
	kill(getppid(), SIGUSR1);
#endif

	pthread_exit(&error);
}


void* threadCyclicDownload(void* ptr)
{
	char url[] = BASE_URL,
		 sys_command[SYSTEM_COMMAND_SIZE];
	Error error;

	signal(SIGTERM, downloaderExit);

#ifdef _WIN32

	if (initWinsockDll()){
		fprintf(stderr, "ERROR:\ninitWinsockDll() failled\n");
		downloaderExit(ERROR_INITWINSOCKDLL);
	}

#endif

	while(true){		// run untill signaled with SIGTERM
#if 1
		if ((io_file = fopen(ZIPPED_FILE_NAME, "w")) == NULL){
			fprintf(stderr,
				"Cannot open file \"%s\"\n", ZIPPED_FILE_NAME);
			downloaderExit(ERROR_OPEN_FILE);
		}

		error = download(io_file, HOST, PORT, updateURL(url));
		if (error != NOT_ERROR){	// error while downloading url
			fprintf(stderr,
				"Error %d occured while downloading URL \"%s\" into file \"%s\"\n",
				error, url, ZIPPED_FILE_NAME);
			downloaderExit(error);
		}

		if (fclose(io_file) == EOF){
			fprintf(stderr,
				"Cannot close file \"%s\"\n", ZIPPED_FILE_NAME);
			downloaderExit(ERROR_CLOSE_FILE);
		}

		// unzip the downloaded file and transform it to a text file
		sprintf(sys_command, "gunzip -q -f %s",ZIPPED_FILE_NAME);
		system(sys_command);
		sprintf(sys_command, "route_btoa -i %s > %s",
				EXTRACTED_FILE_NAME, ASCII_FILE_NAME);
		system(sys_command);
		remove(EXTRACTED_FILE_NAME);
#endif
		if ((io_file = fopen(ASCII_FILE_NAME, "r")) == NULL){
			fprintf(stderr,
				"Cannot open file \"%s\"\n", ASCII_FILE_NAME);
			downloaderExit(ERROR_OPEN_FILE);
		}
	
		pthread_mutex_lock(&trie_locker);

		buildTrie(io_file, trie);

		pthread_mutex_unlock(&trie_locker);

		pthread_cond_broadcast(&startup_event);

		if (fclose(io_file) == EOF){
			fprintf(stderr, "Cannot close file \"%s\"\n", ASCII_FILE_NAME);
			downloaderExit(ERROR_CLOSE_FILE);
		}
//////////////////////////////////////////////////////////
/////// remove(ASCII_FILE_NAME);
//////////////////////////////////////////////////////////

		sleep(download_cycle -
			 ((unsigned)difftime(time(NULL), startup_time) %
			 download_cycle));
	}

	// the thread will never get here
	return NULL;
}
