#include "Global_ds.h"

static void startupDialog()
{
	int value;	// needed to solve signed/unsigned mismatch

	fprintf(stdout, "\nWould you like to change program parameters? [y/n] ");

	if (getc(stdin) != 'y')
		return;

	fprintf(stdout, "\nEnter maximal number of hours to regard a conflict as suspicious\n");
	fprintf(stdout, "(default is %d hours): ", (DEFAULT_SUSPICION_MAX_TIME / (60*60)));
	fscanf(stdin, "%d", &value);
	if (value < 1)
		fprintf(stdout, "Illegal value, using default.\n");
	else{
		fprintf(stdout, "Using: %d hours.\n", value);
		suspicion_max_time = value * 60 * 60;	// transform hours to seconds
	}

	fprintf(stdout, "\nEnter maximal number of conflicts to keep in memory\n");
	fprintf(stdout, "(default is %d conflicts): ", DEFAULT_CONFLICTS_QUEUE_SIZE);
	fscanf(stdin, "%d", &value);
	if (value < 10)
		fprintf(stdout, "Illegal value, using default.\n");
	else{
		conflicts_queue_size = value;
		fprintf(stdout, "Using: %d conflicts.\n", conflicts_queue_size);
	}

	fprintf(stdout, "\nEnter maximal number of conflicts to keep tracking for each IP\n");
	fprintf(stdout, "(default is the last %d conflicts for each IP): ",
			DEFAULT_IP_TRACKING_QUEUE_SIZE);
	fscanf(stdin, "%d", &value);
	if (value < 10)
		fprintf(stdout, "Illegal value, using default.\n");
	else{
		ip_tracking_queue_size = value;
		fprintf(stdout, "Using: %d conflicts.\n", ip_tracking_queue_size);
	}

	fprintf(stdout, "\nEnter number of hours for the downloading cycle\n");
	fprintf(stdout, "(default is %d hours): ", (DEFAULT_DOWNLOAD_CYCLE / (60*60)));
	fscanf(stdin, "%d", &value);
	if (value < 1)
		fprintf(stdout, "Illegal value, using default.\n");
	else{
		fprintf(stdout, "Using: %d hours.\n", value);
		download_cycle = value * 60 * 60;	// transform hours to seconds
	}

	fprintf(stdout, "\nEnter security mode (default is %d): ",
			DEFAULT_SECURITY_MODE);
	fscanf(stdin, "%d", &value);
	if ((value != SECURITY_MIN) &&
		(value != SECURITY_MED) &&
		(value != SECURITY_MAX))
		fprintf(stdout, "Illegal value, using default.\n");
	else{
		security_mode = value;
        fprintf(stdout, "Using: %d security mode.\n", security_mode);
	}
}


Error initializeDS()
{
	startup_time			= time(NULL);

	conflicts_min_counter	= 0;
	conflicts_med_counter	= 0;
	conflicts_max_counter	= 0;

	suspicion_max_time		= DEFAULT_SUSPICION_MAX_TIME;

	conflicts_queue_size	= DEFAULT_CONFLICTS_QUEUE_SIZE;
	ip_tracking_queue_size	= DEFAULT_IP_TRACKING_QUEUE_SIZE;

	download_cycle			= DEFAULT_DOWNLOAD_CYCLE;

	security_mode			= DEFAULT_SECURITY_MODE;

	startupDialog();

	trie					= createTrie();
	if (trie == NULL)
		return ERROR_MEMORY_ALLOCATION;

	conflicts_queue			= createQueue(conflicts_queue_size);
	if (conflicts_queue == NULL){
		freeTrie(trie);
		return ERROR_MEMORY_ALLOCATION;
	}

	pthread_mutex_init(&trie_locker, NULL);
	pthread_cond_init (&startup_event, NULL);

	signal(SIGUSR1, terminateDS);
	signal(SIGTERM, terminateDS);

	fprintf(stdout, "\n\nBuilding up data structures ...\n");
	fprintf(stdout,
		"The process can take a few minutes, please be patient.\n");

	pthread_mutex_lock(&trie_locker);

	if (pthread_create(&downloader_thread, NULL,
					   threadCyclicDownload, NULL)){
		pthread_cond_destroy(&startup_event);
		pthread_mutex_destroy(&trie_locker);
		freeQueue(conflicts_queue);
		freeTrie(trie);
		return ERROR_CREATE_THREAD;
	}

	pthread_cond_wait(&startup_event, &trie_locker);
	pthread_mutex_unlock(&trie_locker);

	return NOT_ERROR;
}


void terminateDS(int n)
{
	FILE *log_file;

#ifndef _WIN32
	pthread_kill(downloader_thread, SIGTERM);
	pthread_join(downloader_thread, NULL);
#endif

	pthread_cond_destroy (&startup_event);
	pthread_mutex_destroy(&trie_locker);

	log_file = fopen(CONFLICTS_LOG_FILE_NAME, "a");

	if (log_file == NULL){
			fprintf(stderr, "Cannot open file \"%s\"\n",
					CONFLICTS_LOG_FILE_NAME);
	}
	else{
		fprintQueue(log_file, conflicts_queue);

		fprintf(log_file, "\n\n********************\n\n");

		fprintf(log_file, "Total number of conflicts since %s",
				ctime(&startup_time));

		fprintf(log_file, "Minimum level conflicts: %d\n",
			conflicts_min_counter);

		fprintf(log_file, "Medium level conflicts:  %d\n",
			conflicts_med_counter);

		fprintf(log_file, "Maximum level conflicts: %d\n",
			conflicts_max_counter);

		if (fclose(log_file) == EOF){
			fprintf(stderr, "Cannot close file \"%s\"\n",
					CONFLICTS_LOG_FILE_NAME);
		}
	}

	freeQueue(conflicts_queue);

	freeTrie(trie);

	exit(n);
}


void registerConflict(Conflict* conflict)
{
	Conflict* old_conflict;

	if (!conflict)
		return;

	old_conflict = push(conflicts_queue, conflict);

	if(old_conflict){
		FILE *log_file = fopen(CONFLICTS_LOG_FILE_NAME, "a");

		if (log_file == NULL){
			fprintf(stderr, "Cannot open file \"%s\"\n",
					CONFLICTS_LOG_FILE_NAME);
		}
		else{
			fprintConflict(log_file, old_conflict);

			fflushConflictsQueue(log_file, conflicts_queue);

			if (fclose(log_file) == EOF){
				fprintf(stderr, "Cannot close file \"%s\"\n",
						CONFLICTS_LOG_FILE_NAME);
			}
		}

		freeConflict(old_conflict);
	}

	switch(conflict->security_level){

	case SECURITY_MIN:
		conflicts_min_counter++;
		break;

	case SECURITY_MED:
		conflicts_med_counter++;
		break;

	case SECURITY_MAX:
		conflicts_max_counter++;
	}
}









#if 0
// older version
// from initializeDS:
	switch(argc){
	case 1:
		input=stdin;
		output=stdout;
		break;
	case 3:
		if ((input = fopen(argv[1], "r")) == NULL)
		{
			fprintf(stderr, "Cannot open file %s\n", argv[1]);
			return ERROR_OPEN_FILE;
		}
		if ((output = fopen(argv[2], "w")) == NULL)
		{
			fprintf(stderr, "Cannot open file %s\n", argv[2]);
			if(fclose(input) == EOF){
				fprintf(stderr, "Cannot close file %s\n", argv[1]);
				return GENERAL_ERROR;
			}
			else
				return ERROR_OPEN_FILE;
		}
		break;
	default:
		fprintf(stderr,
			"USAGE:\nprefix-ownership [<input file> <output file>]\n");
		return ERROR_COMMAND_LINE;
	}

// from terminateDS:
	if(argc == 3 &&
	  (fclose(input) == EOF || fclose(output) == EOF))
			return ERROR_CLOSE_FILE;
#endif