#include "Trie.h"

/* ASList library functions */
/* ------------------------ */

// Note: all the ASList functions are defined as 'static'.

// Allocate memory for an ASList.
// Return: pointer to the new ASList.
static ASList *createASList(const unsigned AS,
							const time_t AS_insert_time)
{
	ASList *asl = (ASList *)malloc(sizeof(ASList));

	if (asl){
		asl->AS_list = NULL;
		insertList(&(asl->AS_list), AS);
		asl->AS_insert_time = AS_insert_time;
	}

    return asl;
}

// Delete all elements in an ASList and free allocated memory.
static void freeASList(ASList *asl)
{
	if (asl){
		freeList(asl->AS_list);
		free(asl);
	}
}

static Conflict *insertASList(ASList **asl,
							  const IP *ip, const unsigned AS,
							  const time_t build_time)
{
	double diff;
	Conflict* new_conflict;

// Not a conflict - the IP is not yet registered in the trie tree
	if (*asl == NULL){
		*asl = createASList(AS, build_time);
		return NULL;
	}

// Not a conflict - the IP is registered with the same AS
	if(findList((*asl)->AS_list, AS))
		return NULL;

	diff = difftime(build_time, (*asl)->AS_insert_time);

// A conflict within a single routing table -
// an IP with more then a single AS was detected;
// Probably legitimate, minimal security level.
	if(diff == 0)
		new_conflict =
			createConflict(ip, AS, copyList((*asl)->AS_list),
						   build_time, (*asl)->AS_insert_time,
						   SECURITY_MIN);

// Two options remain:
// 1)	a conflict that occured after a long period,
//		probably legitimate, medium security level, or:
// 2)	a short-period conflict, suspicious of hijacking,
//		maximal security level
	else{
		new_conflict =
			createConflict(ip, AS, (*asl)->AS_list,
				build_time, (*asl)->AS_insert_time,
				(diff > suspicion_max_time ? SECURITY_MED : SECURITY_MAX));
		(*asl)->AS_list = NULL;
		(*asl)->AS_insert_time = build_time;
	}

	insertList(&((*asl)->AS_list), AS);

	return new_conflict;
}


/* Trie library functions */
/* ---------------------- */

Trie* createTrie()
{
	// allocate memory and initialize it to 0, i.e:
	// both (trie->AS_list[] == NULL) and (trie->children == NULL).
	return (Trie*)calloc(1, sizeof(Trie));
}

void freeTrie(Trie* trie){
	if(trie){
		int i;

		if(trie->AS_array){
			for (i=0; i < MASK_ARRAY_SIZE; ++i)
				freeASList(trie->AS_array[i]);
			free(trie->AS_array);
		}

		if(trie->children){
			for (i=0; i < CHILDREN_ARRAY_SIZE; ++i)
				freeTrie(trie->children[i]);
			free(trie->children);
		}

		free(trie);
	}
}


Conflict *insertPrefix(Trie* trie,
					   const IP *ip, const unsigned AS,
					   const time_t build_time)
{
	short depth,
		  max_depth = (ip->mask - 1)/8 +1,
		  ip_array[IP_SIZE];
	Trie* current = trie;

	IPToShortArray(ip, ip_array);

	for (depth = 0; depth < max_depth; depth++){

		assert(current);
		
		if (current->children == NULL)
			current->children =
				(Trie **)calloc(CHILDREN_ARRAY_SIZE, sizeof(Trie*));
		
		assert(current->children);

		if (current->children[ip_array[depth]] == NULL)
			current->children[ip_array[depth]] = createTrie();

		current = current->children[ip_array[depth]];
	}

	if (current->AS_array == NULL)
		current->AS_array =
			(ASList **)calloc(MASK_ARRAY_SIZE, sizeof(ASList*));
	
	assert(current->AS_array);

	return
		insertASList(&(current->AS_array[ip->mask % MASK_ARRAY_SIZE]),
					 ip, AS, build_time);
}

List* findAS(Trie* trie, const IP *ip)
{
	short depth,
		  ip_array[IP_SIZE],
		  max_depth = (ip->mask - 1)/8 +1,
		  AS_index = ip->mask % MASK_ARRAY_SIZE;
	Trie* current = trie;

	IPToShortArray(ip, ip_array);

	for(depth = 0; depth < max_depth; ++depth){
		if(!current || !current->children)
			return NULL;
		current = current->children[ip_array[depth]];
	}

	if(current && current->AS_array && current->AS_array[AS_index])
		return current->AS_array[AS_index]->AS_list;

	return NULL;
}

void fprintTrie(FILE* output, const Trie* trie)
{
	if(trie){
		int i;

		if (trie->AS_array)
			for(i=0; i < MASK_ARRAY_SIZE; i++)
				if(trie->AS_array[i])
					fprintList(output, trie->AS_array[i]->AS_list);

		if(trie->children){
			for (i=0; i<CHILDREN_ARRAY_SIZE; i++)
				fprintTrie(output, trie->children[i]);
		}
	}
}









#if 0
// older version
static bool traverseTrie(Trie* node, int key)
{
	assert(node);

	if (node->children == NULL)
		node->children =
			(Trie **)calloc(CHILDREN_ARRAY_SIZE, sizeof(Trie*));

	assert(node->children);

	if (node->children[key] == NULL){
		node->children[key] = createTrie();
		return true;
	}

	// else: node->children[key] != NULL
	return false;
}

bool insertPrefix(Trie* trie, short IP[], unsigned AS)
{
	int max_depth	= (IP[IP_SIZE-1]-1)/8 +1,
		mask_index	= IP[IP_SIZE-1] % MASK_ARRAY_SIZE,
		depth;
	bool empty_list_flag,
		 new_node_flag = false,
		 insert_flag;
	Trie* current = trie;

	assert(trie); // assume that the trie was already created

	for (depth=0; depth < max_depth; depth++){
		new_node_flag = traverseTrie(current, IP[depth]);
		current = current->children[IP[depth]];
	}

	if (current->AS_list == NULL)
		current->AS_list =
				(List **)calloc(MASK_ARRAY_SIZE, sizeof(List*));
	
	assert(current->AS_list);

	empty_list_flag = (bool)current->AS_list[mask_index];

	insert_flag = insertList(&(current->AS_list[mask_index]), AS);

	return (new_node_flag || !empty_list_flag || !insert_flag);
}

List* findAS(Trie* trie, short IP[])
{
	int depth, max_depth = (IP[IP_SIZE-1]-1)/8 +1;
	Trie* current = trie;

	for(depth=0; depth < max_depth; ++depth){
		if(!current || !current->children)
			return NULL;
		current = current->children[IP[depth]];
	}

	if(current && current->AS_list)
		return current->AS_list[IP[IP_SIZE-1] % MASK_ARRAY_SIZE];

	return NULL;
}

#endif